/*
 *                                                                         *
 *   Derived from LinuxSampler code                                        *
 *                                                                         *
 *   Copyright (C) 2003, 2004 by Benno Senoner and Christian Schoenebeck   *
 *   Copyright (C) 2005 - 2008 Christian Schoenebeck                       *
 *   Copyright (C) 2013 Grame
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
*/

#include "JackDriverLoader.h"
#include "driver_interface.h"
#include "JackASIODriver.h"
#include "JackEngineControl.h"
#include "JackGraphManager.h"
#include "JackError.h"
#include "JackTime.h"
#include "JackTools.h"
#include "JackCompilerDeps.h"
#include <iostream>
#include <assert.h>

#include <asiosys.h>
#include <asio.h>
#include <asiodrivers.h>
#include <math.h>
#include "iasiothiscallresolver.h"

using namespace std;

namespace Jack
{

extern AsioDrivers* asioDrivers;
extern bool loadAsioDriver(char* name);

static DriverInfo asioDriverInfo = { { 0 }, 0 };

static bool asioDriverOpened = false;
static bool JackASIODriverInstantiated = false;
static String currentAsioDriverName;
static std::vector<String> asioDriverList;
static bool asioDriverListLoaded = false;

// callback prototypes

ASIOCallbacks asioCallbacks;
static JackASIODriver* GlobalJackASIODriverThisPtr;

template<int bitres>
static void floatToASIOSTInt32LSBXX(float* in, void* dest, int numSamples) 
{
    double pos_max_value = (1 << bitres) -1.0;
    double neg_max_value = -pos_max_value;
    int32_t* out = (int32_t*)dest;
    for (int i = 0; i < numSamples; i++) {
        double sample_point = in[i] * pos_max_value;
        if (sample_point < neg_max_value) sample_point = neg_max_value;
        if (sample_point > pos_max_value) sample_point =  pos_max_value;
        out[i] = (int16_t)sample_point;
    }
}

static void floatToASIOSTInt16LSB(float* in, void* dest, int numSamples) 
{
    int16_t* out = (int16_t*)dest;
    for (int i = 0; i < numSamples; i++) {
        float sample_point = in[i] * 32768.0f;
        if (sample_point < -32768.0) sample_point = -32768.0;
        if (sample_point > 32767.0) sample_point = 32767.0;
        out[i] = (int16_t)sample_point;
    }
}

static void floatToASIOSTInt32LSB(float* in, void* dest, int numSamples) 
{
    int32_t* out = (int32_t*)dest;
    for (int i = 0; i < numSamples; i++) {
        double sample_point = in[i] * 2147483648.0;
        if (sample_point < -2147483648.0) sample_point = -2147483648.0;
        if (sample_point > 2147483647.0) sample_point = 2147483647.0;
        out[i] = (int32_t)(sample_point);
    }
}

std::vector<String> getAsioDriverNames();

static bool ASIO_loadAsioDriver(const char* name) 
{
    jack_log("ASIO_loadAsioDriver: trying to load '%s'", name);
#ifdef  WINDOWS
    CoInitialize(0);
#endif
    return loadAsioDriver(const_cast<char*>(name));
}

int ASIO_OpenAndQueryDeviceInfo(String driverName, DriverInfo* driverInfo) 
{
    ASIOSampleRate possibleSampleRates[] =
        { 8000.0, 9600.0, 11025.0, 12000.0, 16000.0, 22050.0, 24000.0,
          32000.0, 44100.0, 48000.0, 88200.0, 96000.0 };

    ASIOError err;

    // Call this function in order to fill the internal vector
    // containing the list of ASIO cards.

    // Since the reading of the list is performed only one time,
    // subsequent calls to this function will be automatically
    // ignored.

    // Calling getAsioDriverNames() is needed in the case of a
    // LinuxSampler client application first creating an ASIO device,
    // which calls ASIO_OpenAndQueryDeviceInfo(), and afterwards
    // retrieve the list of available ASIO cards, since
    // getAsioDriverNames() unloads the current driver and loads the
    // driver 'dummy' this would cause the audio stopping and forcing
    // the create the audio output device again.
    getAsioDriverNames();

    jack_log("ASIO_OpenAndQueryDeviceInfo driverName='%s' current='%s'",
            driverName.c_str(), currentAsioDriverName.c_str());
    if (driverName == currentAsioDriverName) {
        // if the given ASIO driver name == current ASIO driver name
        // and the driver was already opened then do nothing
        if (asioDriverOpened) {
            jack_log("asioDriver already opened, doing nothing");
            return 0;
        }
    } else {
        jack_log("driverName != currentAsioDriverName, new asio driver specified, opening device...");
        // if a new ASIO driver name was specified then first check if
        // we need to close the old one
        if (asioDriverOpened) {
            jack_log("different asioDriver already opened, closing old one");
            asioDriverOpened = false;
            ASIOExit(); // close the old ASIO driver
        }
    }
    currentAsioDriverName = driverName;

    memset(&driverInfo->driverInfo, 0, sizeof(ASIODriverInfo));
    driverInfo->driverInfo.asioVersion = 1;
    driverInfo->driverInfo.sysRef = NULL;

    // MUST BE CHECKED : to force fragments loading on Mac
    // ASIO_loadAsioDriver("dummy");

    jack_log("Before ASIO_loadAsioDriver('%s')", driverName.c_str()));
    if (!ASIO_loadAsioDriver(driverName.c_str())) {
        jack_log("ASIO_OpenAndQueryDeviceInfo could not loadAsioDriver %s", driverName.c_str()));
        return -1;
    }
    jack_log("Before ASIOInit()");
    if ((err = ASIOInit(&driverInfo->driverInfo)) != ASE_OK) {
        jack_log("ASIO_OpenAndQueryDeviceInfo: ASIOInit returned %d for %s",
                int(err), driverName.c_str()));
        return err;
    }
    jack_log("Before ASIOGetChannels()");
    if ((err = ASIOGetChannels(&driverInfo->numInputChannels,
                                     &driverInfo->numOutputChannels)) != ASE_OK) {
        jack_log("ASIO_OpenAndQueryDeviceInfo could not ASIOGetChannels for %s: %d",
                driverName.c_str(), int(err));
        return err;
    }

    jack_log("Before ASIOGetBufferSize()");
    if ((err = ASIOGetBufferSize(&driverInfo->minBufSize,
                                       &driverInfo->maxBufSize,
                                       &driverInfo->preferredBufSize,
                                       &driverInfo->bufGranularity)) != ASE_OK) {
        jack_log("ASIO_OpenAndQueryDeviceInfo could not ASIOGetBufferSize for %s: %d",
                driverName.c_str(), int(err));
        return err;
    }

    jack_log("ASIO_OpenAndQueryDeviceInfo: InputChannels = %ld", driverInfo->numInputChannels));
    jack_log("ASIO_OpenAndQueryDeviceInfo: OutputChannels = %ld", driverInfo->numOutputChannels));

    // Loop through the possible sampling rates and check each to see
    // if the device supports it.
    driverInfo->sampleRates.clear();
    for (int index = 0; index < MAX_NUMSAMPLINGRATES; index++) {
        if (ASIOCanSampleRate(possibleSampleRates[index]) != ASE_NoClock) {
            jack_log("ASIOCanSampleRate: possible sample rate = %ld",
                    (long)possibleSampleRates[index]));
            driverInfo->sampleRates.push_back((int)possibleSampleRates[index]);
        }
    }

    // get the channel infos for each output channel (including sample format)
    for (int i = 0; i < driverInfo->numOutputChannels; i++) {
        driverInfo->channelInfos[i].channel = i;
        driverInfo->channelInfos[i].isInput = ASIOFalse;
        ASIOGetChannelInfo(&driverInfo->channelInfos[i]);
        jack_log("channelInfos[%d].type (sampleformat) = %d",
                i, int(driverInfo->channelInfos[i].type)));
    }

    jack_log("ASIO_OpenAndQueryDeviceInfo: driver opened.");
    asioDriverOpened = true;
    return ASE_OK;
}

std::vector<String> getAsioDriverNames() 
{
    char* names[ASIO_MAX_DEVICE_INFO];
    int numDrivers;

    if (asioDriverListLoaded) {
        jack_log("getAsioDriverNames: ASIO driver list already loaded, doing returning cached list.");
        return asioDriverList;
    }

    // MUST BE CHECKED : to force fragments loading on Mac
    ASIO_loadAsioDriver("dummy");

#if MAC
    numDrivers = asioDrivers->getNumFragments();
#elif WINDOWS
    numDrivers = asioDrivers->asioGetNumDev();
#endif

    for (int i = 0; i < ASIO_MAX_DEVICE_INFO; i++) {
        names[i] = new char[32];
        memset(names[i], 0, 32);
    }

    // Get names of all available ASIO drivers
    asioDrivers->getDriverNames(names, ASIO_MAX_DEVICE_INFO);
    jack_log("getAsioDriverNames: numDrivers = %d", numDrivers);

    for (int i = 0; i < numDrivers; i++) {
        jack_log("ASIO DRIVERLIST: i = %d  name ='%s'", i, names[i]);

#if 1
        asioDriverList.push_back(names[i]);
#else
        // FIXME: We currently try what is the best method to exclude
        // not connected devices or not working drivers. The code
        // below is for testing only, it tries to load each ASIO
        // driver and if it gives an error it is not included in the
        // list of available drivers (asioDriverList).
        if (ASIO_loadAsioDriver(names[i])) {
            ASIOError err;
            if ((err = ASIOInit(&asioDriverInfo.driverInfo)) == ASE_OK) {
                asioDriverList.push_back(names[i]);
            } else {
                jack_log("getDriverList: ASIOInit of driver %s gave Error %d! ignoring it.",
                        names[i], int(err));
            }
        }
        else {
            jack_log("getDriverList: load driver %s failed! ignoring it.", names[i]);
        }
        // FIXME: we need to check this ASIOExit is needed (gave a
        // crash on a ASIO ADSP24(WDM) card so we commented it out)
        // ASIOExit();
#endif
        //currentAsioDriverName = "";
    }

    for (int i = 0; i < ASIO_MAX_DEVICE_INFO; i++) {
        delete[] names[i];
    }

    jack_log("getAsioDriverNames: returning from function. asioDriverList.size() = %d",
            int(asioDriverList.size()));
    asioDriverListLoaded = true;
    return asioDriverList;
}

// currently not used
unsigned long get_sys_reference_time() 
{
    // get the system reference time
#if WINDOWS
    return timeGetTime();
#elif MAC
    static const double twoRaisedTo32 = 4294967296.;
    UnsignedWide ys;
    Microseconds(&ys);
    double r = ((double)ys.hi * twoRaisedTo32 + (double)ys.lo);
    return (unsigned long)(r / 1000.);
#endif
}

//----------------------------------------------------------------------------------
// conversion from 64 bit ASIOSample/ASIOTimeStamp to double float
#if NATIVE_INT64
#define ASIO64toDouble(a)  (a)
#else
const double twoRaisedTo32 = 4294967296.;
#define ASIO64toDouble(a)  ((a).lo + (a).hi * twoRaisedTo32)
#endif

ASIOTime* JackASIODriver::bufferSwitchTimeInfo(ASIOTime *timeInfo, long index, ASIOBool processNow) 
{
    static long processedSamples = 0;
    // store the timeInfo for later use
    asioDriverInfo.tInfo = *timeInfo;

    // get the time stamp of the buffer, not necessary if no
    // synchronization to other media is required
    if (timeInfo->timeInfo.flags & kSystemTimeValid) {
        asioDriverInfo.nanoSeconds = ASIO64toDouble(timeInfo->timeInfo.systemTime);
    } else {
        asioDriverInfo.nanoSeconds = 0;
    }

    if (timeInfo->timeInfo.flags & kSamplePositionValid) {
        asioDriverInfo.samples = ASIO64toDouble(timeInfo->timeInfo.samplePosition);
    } else {
        asioDriverInfo.samples = 0;
    }

    if (timeInfo->timeCode.flags & kTcValid) {
        asioDriverInfo.tcSamples = ASIO64toDouble(timeInfo->timeCode.timeCodeSamples);
    } else {
        asioDriverInfo.tcSamples = 0;
    }

    // TODO: ignore for now. get the system reference time
    // asioDriverInfo.sysRefTime = get_sys_reference_time();

    // buffer size in samples
    long buffSize = asioDriverInfo.chosenBufSize;

    // tell JACK to  render a fragment of buffSize samples
    GlobalJackASIODriverThisPtr->RenderAudio(buffSize);

    // now write and convert the samples to the ASIO buffer
    for (int i = 0; i < asioDriverInfo.numOutputBuffers; i++)
    {
        // do processing for the outputs only
        switch (asioDriverInfo.channelInfos[i].type)
        {
        case ASIOSTInt16LSB:
            floatToASIOSTInt16LSB(GlobalJackASIODriverThisPtr->Channels[i]->Buffer(),
                                  (int16_t*)asioDriverInfo.bufferInfos[i].buffers[index], buffSize);
            break;
        case ASIOSTInt24LSB:   // used for 20 bits as well
            memset(asioDriverInfo.bufferInfos[i].buffers[index], 0, buffSize * 3);
            break;
        case ASIOSTInt32LSB:
            floatToASIOSTInt32LSB(GlobalJackASIODriverThisPtr->Channels[i]->Buffer(),
                                  (int32_t*)asioDriverInfo.bufferInfos[i].buffers[index], buffSize);
            break;
        case ASIOSTFloat32LSB: // IEEE 754 32 bit float, as found on Intel x86 architecture
            throw AudioOutputException("ASIO Error: ASIOSTFloat32LSB not yet supported! report to JACK developers.");
            break;
        case ASIOSTFloat64LSB: // IEEE 754 64 bit double float, as found on Intel x86 architecture
            throw AudioOutputException("ASIO Error: ASIOSTFloat64LSB not yet supported! report to JACK developers.");
            break;

            // These are used for 32 bit data buffer, with different
            // alignment of the data inside. 32 bit PCI bus systems
            // can more easily used with these.

        case ASIOSTInt32LSB16: // 32 bit data with 16 bit alignment
            floatToASIOSTInt32LSBXX<16>(GlobalJackASIODriverThisPtr->Channels[i]->Buffer(),
                                        (int32_t*)asioDriverInfo.bufferInfos[i].buffers[index],
                                        buffSize);
            break;
        case ASIOSTInt32LSB18: // 32 bit data with 18 bit alignment
            floatToASIOSTInt32LSBXX<18>(GlobalJackASIODriverThisPtr->Channels[i]->Buffer(),
                                        (int32_t*)asioDriverInfo.bufferInfos[i].buffers[index],
                                        buffSize);
            break;
        case ASIOSTInt32LSB20: // 32 bit data with 20 bit alignment
            floatToASIOSTInt32LSBXX<20>(GlobalJackASIODriverThisPtr->Channels[i]->Buffer(),
                                        (int32_t*)asioDriverInfo.bufferInfos[i].buffers[index],
                                        buffSize);
            break;
        case ASIOSTInt32LSB24: // 32 bit data with 24 bit alignment
            floatToASIOSTInt32LSBXX<24>(GlobalJackASIODriverThisPtr->Channels[i]->Buffer(),
                                        (int32_t*)asioDriverInfo.bufferInfos[i].buffers[index],
                                        buffSize);
            break;
        case ASIOSTInt16MSB:
            throw AudioOutputException("ASIO Error: ASIOSTInt16MSB not yet supported! report to JACK developers.");
            break;
        case ASIOSTInt24MSB:   // used for 20 bits as well
            throw AudioOutputException("ASIO Error: ASIOSTInt24MSB not yet supported! report to JACK developers.");
            break;
        case ASIOSTInt32MSB:
            throw AudioOutputException("ASIO Error: ASIOSTInt32MSB not yet supported! report to JACK developers.");
            break;
        case ASIOSTFloat32MSB: // IEEE 754 32 bit float, as found on bigendian architecture
            throw AudioOutputException("ASIO Error: ASIOSTFloat32MSB not yet supported! report to JACK developers.");
            break;
        case ASIOSTFloat64MSB: // IEEE 754 64 bit double float, as found on bigendian architecture
            throw AudioOutputException("ASIO Error: ASIOSTFloat64MSB not yet supported! report to JACK developers.");
            break;

            // These are used for 32 bit data buffer, with different
            // alignment of the data inside.  32 bit PCI bus systems
            // can more easily used with these.

        case ASIOSTInt32MSB16: // 32 bit data with 18 bit alignment
            throw AudioOutputException("ASIO Error: ASIOSTInt32MSB16 not yet supported! report to JACK developers.");
            break;
        case ASIOSTInt32MSB18: // 32 bit data with 18 bit alignment
            throw AudioOutputException("ASIO Error: ASIOSTInt32MSB18 not yet supported! report to JACK developers.");
            break;
        case ASIOSTInt32MSB20: // 32 bit data with 20 bit alignment
            throw AudioOutputException("ASIO Error: ASIOSTInt32MSB20 not yet supported! report to JACK developers.");
            break;
        case ASIOSTInt32MSB24: // 32 bit data with 24 bit alignment
            throw AudioOutputException("ASIO Error: ASIOSTInt32MSB24 not yet supported! report to JACK developers.");
            break;
        default:
            throw AudioOutputException("ASIO Error: unknown ASIO sample format. report error.");
            break;
        }
    }

    // finally if the driver supports the ASIOOutputReady()
    // optimization, do it here, all data are in place
    if (asioDriverInfo.ASIOOutputReadySupported) {
        ASIOOutputReady();
    }

    processedSamples += buffSize;
    return 0L;
}

void JackASIODriver::bufferSwitch(long index, ASIOBool processNow)
{
    // The actual processing callback.

    // As this is a "back door" into the bufferSwitchTimeInfo, a
    // timeInfo needs to be created, though it will only set the
    // timeInfo.samplePosition and timeInfo.systemTime fields and the
    // according flags.
    ASIOTime timeInfo;
    memset(&timeInfo, 0, sizeof(timeInfo));

    // Get the time stamp of the buffer. Not necessary if no
    // synchronization to other media is required.
    if (ASIOGetSamplePosition(&timeInfo.timeInfo.samplePosition,
                              &timeInfo.timeInfo.systemTime) == ASE_OK) {
        timeInfo.timeInfo.flags = kSystemTimeValid | kSamplePositionValid;
    }

    bufferSwitchTimeInfo(&timeInfo, index, processNow);
}

void sampleRateChanged(ASIOSampleRate sRate) 
{
    // Do whatever you need to do if the sample rate changed.
    // Usually this only happens during external sync.
    // Audio processing is not stopped by the driver, actual sample rate
    // might not have even changed, maybe only the sample rate status of an
    // AES/EBU or S/PDIF digital input at the audio device.
    // You might have to update time/sample related conversion routines, etc.
}

long asioMessages(long selector, long value, void* message, double* opt) 
{
    jack_log("asioMessages selector = %ld value = %ld", selector, value);
    // currently the parameters "value", "message" and "opt" are not used.
    long ret = 0;
    switch (selector)
    {
    case kAsioSelectorSupported:
        if (value == kAsioResetRequest ||
            value == kAsioEngineVersion ||
            value == kAsioResyncRequest ||
            value == kAsioLatenciesChanged ||
            // the following three were added for ASIO 2.0, we don't necessarily have to support them
            value == kAsioSupportsTimeInfo ||
            value == kAsioSupportsTimeCode ||
            value == kAsioSupportsInputMonitor)
            ret = 1L;
        break;
    case kAsioResetRequest:
        // Defer the task and perform the reset of the driver during the next "safe" situation.
        // You cannot reset the driver right now, as this code is called from the driver.
        // Reset the driver is done by completely destruct is. I.e. ASIOStop(),
        // ASIODisposeBuffers(), Destruction
        // Afterwards you initialize the driver again.

        //GlobalJackASIODriverThisPtr->asioDriverInfo.stopped; // In this sample the processing will just stop
        ret = 1L;
        break;
    case kAsioResyncRequest:
        // This informs the application, that the driver encountered some non fatal data loss.
        // It is used for synchronization purposes of different media.
        // Added mainly to work around the Win16Mutex problems in Windows 95/98 with the
        // Windows Multimedia system, which could loose data because the Mutex was hold too long
        // by another thread.
        // However a driver can issue it in other situations, too.
        ret = 1L;
        break;
    case kAsioLatenciesChanged:
        // This will inform the host application that the drivers were latencies changed.
        // Beware, it this does not mean that the buffer sizes have changed!
        // You might need to update internal delay data.
        ret = 1L;
        break;
    case kAsioEngineVersion:
        // Return the supported ASIO version of the host application.
        // If a host applications does not implement this selector, ASIO 1.0 is assumed
        // by the driver
        ret = 2L;
        break;
    case kAsioSupportsTimeInfo:
        // Informs the driver whether the asioCallbacks.bufferSwitchTimeInfo() callback
        // is supported.
        // For compatibility with ASIO 1.0 drivers the host application should always support
        // the "old" bufferSwitch method, too.
        ret = 1;
        break;
    case kAsioSupportsTimeCode:
        // Informs the driver whether application is interested in time code info.
        // If an application does not need to know about time code, the driver has less work
        // to do.
        ret = 0;
        break;
    }
    return ret;
}

int JackASIODriver::Read()
{
    // TODO
    return 0;
}

int JackASIODriver::Write()
{
    // TODO
    return 0;
}

void JackASIODriver::UpdateLatencies()
{
    jack_latency_range_t input_range;
    jack_latency_range_t output_range;
    jack_latency_range_t monitor_range;

    /*
    const PaStreamInfo* info = Pa_GetStreamInfo(fStream);
    assert(info);

    for (int i = 0; i < fCaptureChannels; i++) {
        input_range.max = input_range.min = fEngineControl->fBufferSize + (info->inputLatency * fEngineControl->fSampleRate) + fCaptureLatency;
        fGraphManager->GetPort(fCapturePortList[i])->SetLatencyRange(JackCaptureLatency, &input_range);
    }

    for (int i = 0; i < fPlaybackChannels; i++) {
        output_range.max = output_range.min = (info->outputLatency * fEngineControl->fSampleRate) + fPlaybackLatency;
        if (fEngineControl->fSyncMode) {
            output_range.max = output_range.min += fEngineControl->fBufferSize;
        } else {
            output_range.max = output_range.min += fEngineControl->fBufferSize * 2;
        }
        fGraphManager->GetPort(fPlaybackPortList[i])->SetLatencyRange(JackPlaybackLatency, &output_range);
        if (fWithMonitorPorts) {
            monitor_range.min = monitor_range.max = fEngineControl->fBufferSize;
            fGraphManager->GetPort(fMonitorPortList[i])->SetLatencyRange(JackCaptureLatency, &monitor_range);
        }
    }
    */
}

int JackASIODriver::Open(jack_nframes_t buffer_size,
                              jack_nframes_t samplerate,
                              bool capturing,
                              bool playing,
                              int inchannels,
                              int outchannels,
                              bool monitor,
                              const char* capture_driver_uid,
                              const char* playback_driver_uid,
                              jack_nframes_t capture_latency,
                              jack_nframes_t playback_latency)
{

    return 0;

error:
    JackAudioDriver::Close();
    jack_error("Can't open default ASIO device");
    return -1;
}

int JackASIODriver::Close()
{
    // Generic audio driver close
    jack_log("JackASIODriver::Close");
    JackAudioDriver::Close();
    ASIOError err = ASIOExit();
    if (err != ASE_OK) {
        jack_error("ASIOExit error = %d", int(err));
    }
    return (err != ASE_OK) ? -1 : 0;
}

int JackASIODriver::Attach()
{
    /*
    if (JackAudioDriver::Attach() == 0) {

        const char* alias;

        if (fInputDevice != paNoDevice && fPaDevices->GetHostFromDevice(fInputDevice) == "ASIO") {
            for (int i = 0; i < fCaptureChannels; i++) {
                if (PaAsio_GetInputChannelName(fInputDevice, i, &alias) == paNoError) {
                    JackPort* port = fGraphManager->GetPort(fCapturePortList[i]);
                    port->SetAlias(alias);
                }
            }
        }

        if (fOutputDevice != paNoDevice && fPaDevices->GetHostFromDevice(fOutputDevice) == "ASIO") {
            for (int i = 0; i < fPlaybackChannels; i++) {
                if (PaAsio_GetOutputChannelName(fOutputDevice, i, &alias) == paNoError) {
                    JackPort* port = fGraphManager->GetPort(fPlaybackPortList[i]);
                    port->SetAlias(alias);
                }
            }
        }

        return 0;

    } else {
        return -1;
    }
    */
    return 0;
}

int JackASIODriver::Start()
{
    jack_log("JackASIODriver::Start");
    if (JackAudioDriver::Start() >= 0) {
        ASIOError err;
        if ((err = ASIOStart()) == ASE_OK) {
            return 0;
        } 
        jack_error("ASIOStart error = %d", int(err));
        JackAudioDriver::Stop();
    }
    return -1;
}

int JackASIODriver::Stop()
{
    jack_log("JackASIODriver::Stop");
    ASIOError err;
    if ((err = ASIOStop()) != ASE_OK) {
       jack_error("ASIOStop error = %d", int(err));
    } 
    if (JackAudioDriver::Stop() < 0) {
        return -1;
    } else {
        return (err == ASE_OK) ? 0 : -1;
    }
}

int JackASIODriver::SetBufferSize(jack_nframes_t buffer_size)
{
/*
    PaError err;

    if (fStream && (err = Pa_CloseStream(fStream)) != paNoError) {
        jack_error("Pa_CloseStream error = %s", Pa_GetErrorText(err));
        goto error;
    }
    
    // It seems that some ASIO drivers (like ASIO4All) needs this to restart correctly;
    delete fPaDevices;
    fPaDevices = new PortAudioDevices();
  
    err = OpenStream(buffer_size);
    if (err != paNoError) {
        jack_error("Pa_OpenStream error %d = %s", err, Pa_GetErrorText(err));
        goto error;
    } else {
        JackAudioDriver::SetBufferSize(buffer_size); // Generic change, never fails
        return 0;
    }
    
error:
    fStream = NULL;
    return -1;
*/
    return 0;
}

} // end of namespace

#ifdef __cplusplus
extern "C"
{
#endif

#include "JackCompilerDeps.h"

    SERVER_EXPORT jack_driver_desc_t* driver_get_descriptor()
    {
        jack_driver_desc_t * desc;
        jack_driver_desc_filler_t filler;
        jack_driver_param_value_t value;

        desc = jack_driver_descriptor_construct("portaudio", JackDriverMaster, "ASIO API based audio backend", &filler);

        value.ui = 0;
        jack_driver_descriptor_add_parameter(desc, &filler, "channels", 'c', JackDriverParamUInt, &value, NULL, "Maximum number of channels", NULL);
        jack_driver_descriptor_add_parameter(desc, &filler, "inchannels", 'i', JackDriverParamUInt, &value, NULL, "Maximum number of input channels", NULL);
        jack_driver_descriptor_add_parameter(desc, &filler, "outchannels", 'o', JackDriverParamUInt, &value, NULL, "Maximum number of output channels", NULL);

        jack_driver_descriptor_add_parameter(desc, &filler, "capture", 'C', JackDriverParamString, &value, NULL, "Provide capture ports. Optionally set PortAudio device name", NULL);

        jack_driver_descriptor_add_parameter(desc, &filler, "playback", 'P', JackDriverParamString, &value, NULL, "Provide playback ports. Optionally set PortAudio device name", NULL);

        value.i = 0;
        jack_driver_descriptor_add_parameter(desc, &filler, "monitor", 'm', JackDriverParamBool, &value, NULL, "Provide monitor ports for the output", NULL);

        value.i = TRUE;
        jack_driver_descriptor_add_parameter(desc, &filler, "duplex", 'D', JackDriverParamBool, &value, NULL, "Provide both capture and playback ports", NULL);

        value.ui = 44100U;
        jack_driver_descriptor_add_parameter(desc, &filler, "rate", 'r', JackDriverParamUInt, &value, NULL, "Sample rate", NULL);

        value.ui = 512U;
        jack_driver_descriptor_add_parameter(desc, &filler, "period", 'p', JackDriverParamUInt, &value, NULL, "Frames per period", "Frames per period. If 0 and ASIO driver, will take preferred value");

        jack_driver_descriptor_add_parameter(desc, &filler, "device", 'd', JackDriverParamString, &value, NULL, "PortAudio device name", NULL);

        value.ui = 0;
        jack_driver_descriptor_add_parameter(desc, &filler, "input-latency", 'I', JackDriverParamUInt, &value, NULL, "Extra input latency", NULL);
        jack_driver_descriptor_add_parameter(desc, &filler, "output-latency", 'O', JackDriverParamUInt, &value, NULL, "Extra output latency", NULL);

        value.i = TRUE;
        jack_driver_descriptor_add_parameter(desc, &filler, "list-devices", 'l', JackDriverParamBool, &value, NULL, "Display available ASIO devices", NULL);

        return desc;
    }

    SERVER_EXPORT Jack::JackDriverClientInterface* driver_initialize(Jack::JackLockedEngine* engine, Jack::JackSynchro* table, const JSList* params)
    {
        jack_nframes_t srate = 44100;
        jack_nframes_t frames_per_interrupt = 512;
        const char* capture_pcm_name = "";
        const char* playback_pcm_name = "";
        bool capture = false;
        bool playback = false;
        int chan_in = 0;
        int chan_out = 0;
        bool monitor = false;
        const JSList *node;
        const jack_driver_param_t *param;
        jack_nframes_t systemic_input_latency = 0;
        jack_nframes_t systemic_output_latency = 0;
    
        for (node = params; node; node = jack_slist_next(node))
        {
            param = (const jack_driver_param_t *) node->data;

            switch (param->character) {

            case 'd':
                capture_pcm_name = param->value.str;
                playback_pcm_name = param->value.str;
                break;

            case 'D':
                capture = true;
                playback = true;
                break;

            case 'c':
                chan_in = chan_out = (int)param->value.ui;
                break;

            case 'i':
                chan_in = (int)param->value.ui;
                break;

            case 'o':
                chan_out = (int)param->value.ui;
                break;

            case 'C':
                capture = true;
                if (strcmp(param->value.str, "none") != 0) {
                    capture_pcm_name = param->value.str;
                }
                break;

            case 'P':
                playback = TRUE;
                if (strcmp(param->value.str, "none") != 0) {
                    playback_pcm_name = param->value.str;
                }
                break;

            case 'm':
                monitor = param->value.i;
                break;

            case 'r':
                srate = param->value.ui;
                break;

            case 'p':
                frames_per_interrupt = (unsigned int)param->value.ui;
                break;

            case 'I':
                systemic_input_latency = param->value.ui;
                break;

            case 'O':
                systemic_output_latency = param->value.ui;
                break;
            
            case 'l':
                // Stops the server in this case
                return NULL;
            }
            
        }

        // duplex is the default
        if (!capture && !playback) {
            capture = true;
            playback = true;
        }

        Jack::JackDriverClientInterface* driver = new Jack::JackASIODriver("system", "asio", engine, table);
        if (driver->Open(frames_per_interrupt, srate, capture, playback,
            chan_in, chan_out, monitor, capture_pcm_name,
            playback_pcm_name, systemic_input_latency,
            systemic_output_latency) == 0) {
            return driver;
        } else {
            delete driver;
            return NULL;
        }
    }

#ifdef __cplusplus
}
#endif
