/*
 *                                                                         *
 *   Derived from LinuxSampler code                                        *
 *                                                                         *
 *   Copyright (C) 2003, 2004 by Benno Senoner and Christian Schoenebeck   *
 *   Copyright (C) 2005 - 2008 Christian Schoenebeck                       *
 *   Copyright (C) 2013 Grame
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
*/

#ifndef __JackASIODriver__
#define __JackASIODriver__

#include "JackAudioDriver.h"
#include "JackMMCSS.h"

namespace Jack
{

#define kMaxInputChannels 32
#define kMaxOutputChannels 32
#define ASIO_MAX_DEVICE_INFO 32

#define NUM_STANDARDSAMPLINGRATES 3  // 11.025, 22.05, 44.1
#define NUM_CUSTOMSAMPLINGRATES   9  // must be the same number of elements as in the array below
#define MAX_NUMSAMPLINGRATES  (NUM_STANDARDSAMPLINGRATES + NUM_CUSTOMSAMPLINGRATES)

// internal data storage
struct DriverInfo {
    // ASIOInit()
    ASIODriverInfo driverInfo;

    // ASIOGetChannels()
    long           numInputChannels;
    long           numOutputChannels;

    // ASIOGetBufferSize()
    long           minBufSize;
    long           maxBufSize;
    long           preferredBufSize;
    long           bufGranularity;
    long           chosenBufSize;

    // ASIOGetSampleRate()
    ASIOSampleRate sampleRate;
    std::vector<int> sampleRates;

    // ASIOOutputReady()
    bool           ASIOOutputReadySupported;

    // ASIOGetLatencies ()
    long           inputLatency;
    long           outputLatency;

    // ASIOCreateBuffers ()
    long           numInputBuffers;  // becomes number of actual created input buffers
    long           numOutputBuffers; // becomes number of actual created output buffers
    ASIOBufferInfo bufferInfos[kMaxInputChannels + kMaxOutputChannels];

    // ASIOGetChannelInfo()
    ASIOChannelInfo channelInfos[kMaxInputChannels + kMaxOutputChannels];
    // The above two arrays share the same indexing, as the data in them are linked together

    // Information from ASIOGetSamplePosition()
    // data is converted to double floats for easier use, however 64 bit integer can be used, too
    double         nanoSeconds;
    double         samples;
    double         tcSamples;   // time code samples

    // bufferSwitchTimeInfo()
    ASIOTime       tInfo;       // time info state
    unsigned long  sysRefTime;  // system reference time, when bufferSwitch() was called

    // Signal the end of processing in this example
    bool           stopped;
};


/*!
\brief The ASIO driver.
*/

class JackASIODriver : public JackMMCSS, public JackAudioDriver
{

    private:

        static ASIOTime* bufferSwitchTimeInfo(ASIOTime *timeInfo, long index, ASIOBool processNow);
        static void bufferSwitch(long index, ASIOBool processNow);

    public:

        JackASIODriver(const char* name, const char* alias, JackLockedEngine* engine, JackSynchro* table)
                : JackMMCSS(), JackAudioDriver(name, alias, engine, table)
        {}

        virtual ~JackASIODriver()
        {}

        int Open(jack_nframes_t buffe_size,
                 jack_nframes_t samplerate,
                 bool capturing,
                 bool playing,
                 int chan_in,
                 int chan_out,
                 bool monitor,
                 const char* capture_driver_name,
                 const char* playback_driver_name,
                 jack_nframes_t capture_latency,
                 jack_nframes_t playback_latency);

        int Close();

        int Attach();

        int Start();
        int Stop();

        int Read();
        int Write();

        // BufferSize can be changed
        bool IsFixedBufferSize()
        {
            return false;
        }

        int SetBufferSize(jack_nframes_t buffer_size);
};

} // end of namespace

#endif
